# Tool that invokes a script in a project local bin directory.
#
# The scripts are called with the project path as the first argument.
import logging

import menhir.tool as mtool
from menhir.tool_utils import OK, FAIL, NOTHING_TO_DO, tool_env, working_dir

log = logging.getLogger(__name__)


class Script():
    pass


def tool():
    return Script()


@mtool.name.register(Script)
def name(arg):
    return "script"


@mtool.is_using_tool.register(Script)
def is_using_tool(tool, path):
    from os.path import exists, join
    return exists(join(path, 'bin'))


@mtool.dependencies.register(Script)
def dependencies(tool, path):
    return []


@mtool.execute_build_phase.register(Script)
def execute_build_phase(
        tool, path, info, phase_name, args,
        only_if_changed=False,
        verbose=False,
):
    """Execute a build phase."""
    import os
    import os.path
    import subprocess

    with working_dir(path):
        script_file = os.path.join('bin', phase_name)
        if os.access(script_file, os.X_OK):
            if verbose:
                print('Running script bin/%s in %s' % (phase_name, path))

            if only_if_changed and not (
                    info.get('changed') or
                    info.get('changed_dependents')
            ):
                return OK

            env = tool_env()
            if info.get('changed'):
                env['MENHIR_CHANGED'] = "1"
            if info.get('changed_dependents'):
                env['MENHIR_CHANGED_DEPENDENTS'] = "1"
            if info.get('changed_dependees'):
                env['MENHIR_CHANGED_DEPENDEES'] = "1"

            result = subprocess.call(
                [script_file] + args,
                env=env,
            )
            log.debug('Script %s result: %s', phase_name, result)
            if result:
                return FAIL
            return OK
        else:
            log.debug('No script for build phase %s', phase_name)
            return NOTHING_TO_DO



# pytest based tool for menhir
import logging

import menhir.tool as mtool
from menhir.tool_utils import (
    OK,
    FAIL,
    NOTHING_TO_DO,
    package_script,
    tool_env,
    working_dir,
)

log = logging.getLogger(__name__)


class Pytest():
    pass


def tool():
    return Pytest()


@mtool.name.register(Pytest)
def name(arg):
    return "pytest"


@mtool.is_using_tool.register(Pytest)
def is_using_tool(tool, path):
    from os.path import exists, join
    path = join(path, 'setup.cfg')
    if exists(path):
        with open(path, "r") as file:
            data = file.read()
            return 'pytest' in data


@mtool.dependencies.register(Pytest)
def dependencies(tool, path):
    return []


@mtool.execute_build_phase.register(Pytest)
def execute_build_phase(
        tool, path, info, phase_name, args,
        only_if_changed=False,
        verbose=False,
):
    """Execute a build phase."""
    import subprocess
    log.info('pytest: %s %s', phase_name, info)
    if phase_name == 'test' and (
            info.get('changed') or info.get('changed_dependents')
    ):
        if verbose:
            print('Running pytest in %s' % path)
        env = tool_env()
        with package_script("/tools/pytest/test.sh") as f:
            with working_dir(path):
                res = subprocess.call([f.name] + args, env=env)
                if res:
                    return FAIL
                return OK
    else:
        return NOTHING_TO_DO



# menhir configuration
import logging

log = logging.getLogger(__name__)
ROOT_FILE = 'menhir_root.yaml'


def find_root():
    """Based on the current directory, look for a .menhir_root.yaml file."""
    from .fileutils import find_file_in_parents
    root = find_file_in_parents(ROOT_FILE)
    if root is None:
        raise FileNotFoundError(
            'No root file found in parent directories',
            ROOT_FILE
        )
    return root


def root_config(root_file=ROOT_FILE):
    """Return the root menhir configuration."""
    from os.path import exists
    import yaml

    if exists(root_file):
        with open(root_file, 'r') as stream:
            config = yaml.load(stream)
    else:
        config = {
            'tools': default_tools(),
        }

    return parse_config(config)


def default_tools():
    from menhir.tools import all_tools
    return all_tools


def parse_config(config):
    """Parse the root menhir configuration."""
    import os
    from menhir import gitutils
    from menhir.graph import root_nodes
    tools = config.get('tools')
    if tools is None:
        raise Exception(
            'missing_configuration',
            'No tools specified in menhir root config')
    tool_impls = list(map(load_tool, tools))

    dirs = {}
    graph = {}
    roots = ['.']
    while roots:
        d = roots.pop()
        if d not in dirs:
            log.debug('Adding directory: %s', d)
            ds, g, e = dirs_and_graph(d, tool_impls)
            dirs.update(ds)
            graph.update(g)
            e = e.difference(set(dirs.keys()))
            log.debug('Adding discovered nodes: %s', e)
            roots.extend(list(e))

    roots = root_nodes(graph, dirs)
    log.debug('root_projects %s', roots)

    git_root = gitutils.find_root()
    cwd = os.getcwd()
    git_prefix = cwd[len(git_root)+1:]

    config['tool_impls'] = tool_impls
    config['project_infos'] = dirs
    config['dependencies'] = graph
    config['root_projects'] = roots
    config['git_root'] = git_root
    config['git_prefix'] = git_prefix
    return config


def dirs_and_graph(root, tool_impls):
    from menhir.graph import leaf_nodes
    dirs = project_infos(root, tool_impls)
    log.debug('project_infos %s', dirs)
    graph = project_graph(dirs)
    leaves = leaf_nodes(graph)
    extra_nodes = leaves.difference(dirs)
    return dirs, graph, extra_nodes


def load_tool(tool_name):
    """Load the tool implementation for the given tool name."""
    tool = \
        require_tool(tool_name) or \
        require_tool('menhir.tools.%s' % tool_name)
    if not tool:
        raise Exception('configuration_error', 'Tool not found %s', tool_name)
    return tool


def require_tool(tool_path):
    from importlib import import_module
    try:
        m = import_module(tool_path)
        return m.tool()
    except Exception as e:
        log.debug('Failed to import: %s, "%s"', tool_path, e)


def project_infos(base_dir, tools):
    """Return a map of project root path to project info dict."""
    from os import walk
    from os.path import normpath
    import menhir.tool as mtool
    roots = {}
    for root, subdirs, files in walk(base_dir):
        used = list(filter(lambda t: mtool.is_using_tool(t, root), tools))
        if len(used):
            roots[normpath(root)] = {'tools': used}
    return roots


def project_graph(roots):
    """Return a project dependency graph."""
    from functools import reduce
    import menhir.tool as mtools

    return {
        root: reduce(
            lambda x, y: x.union(y or set()),
            [
                set(mtools.dependencies(tool, root))
                for tool in info['tools']
            ],
            set()
        )
        for root, info in roots.items()
        }



# Tool interface for tool implementations
try:
    from functools import singledispatch
except:
    from singledispatch import singledispatch


def missing_impl(tool, f):
    raise Exception(
        'incorrect tool implementation',
        'Tool implementation %s is missing a "%s" implementation' % (
            type(tool),
            f,
        )
    )


@singledispatch
def name(tool):
    missing_impl(tool, name.__name__)


@singledispatch
def is_using_tool(tool, path):
    """Predicate to test if the path contains a project using the tool."""
    missing_impl(tool, is_using_tool.__name__)


@singledispatch
def dependencies(tool, path):
    """Return a list of dependency prefixes for the project at path."""
    missing_impl(tool, dependencies.__name__)


@singledispatch
def execute_build_phase(tool, path, info, phase_name, args, **kwargs):
    """Execute a build phase."""
    missing_impl(tool, dependencies.__name__)



from __future__ import print_function  # NOQA

import logging

log = logging.getLogger(__name__)


def main():
    args = parser().parse_args()
    logging.basicConfig(level=args.log_level)
    if args.phase:
        exec_phase(
            args.phase,
            args.args,
            args.only_if_changed,
            args.verbose,
            args.commit,
            args.circleci,
        )


def parser():
    import argparse
    parser = argparse.ArgumentParser(description='Menhir monorepo build tool.')
    parser.add_argument("--commit")
    parser.add_argument("-v", "--verbose", default=False,
                        action='store_const', const=True)
    parser.add_argument("--circleci", default=False,
                        action='store_const', const=True)
    parser.add_argument("--log-level", default='WARNING')
    subparsers = parser.add_subparsers(help='sub-commands')
    parser_exec = subparsers.add_parser('exec', help='execute a build phase')
    parser_exec.add_argument("--only-if-changed", default=False,
                             action='store_const', const=True)
    parser_exec.add_argument('phase')
    parser_exec.add_argument('args', nargs='*')
    return parser


def exec_phase(phase, args, only_if_changed, verbose, commit, circleci):
    import sys
    from menhir import (
        apply_phase,
        update_for_changed_files,
        root_config,
    )

    config = root_config()

    project_infos = config['project_infos']
    dependencies = config['dependencies']
    root_projects = config['root_projects']
    git_prefix = config['git_prefix']

    if verbose:
        print_project_infos(project_infos)
        print_dependencies(dependencies)
        print_git_info(config['git_root'], git_prefix)

    files = changed_files(config, circleci, commit, verbose)
    if verbose:
        print_changed_files(files)

    changed_projects = update_for_changed_files(
        dependencies, project_infos, root_projects, git_prefix, files
    )
    if verbose:
        print_changed_projects(changed_projects)
        print_project_info(project_infos)

    res = apply_phase(
        dependencies,
        project_infos,
        root_projects,
        phase,
        args,
        only_if_changed=only_if_changed,
        verbose=verbose,
    )

    if res.get('failed'):
        print('Build phase "%s" failed' % phase, file=sys.stderr)
        sys.exit(1)
    else:
        sys.exit(0)


def print_project_infos(project_infos):
    print('Project dirs:')
    for p in project_infos:
        print('    ', p)


def print_dependencies(dependencies):
    print('Dependencies Graph:')
    for p in dependencies.items():
        print('    ', p)


def print_git_info(git_root, git_prefix):
    print("Git root:", git_root)
    print("Git prefix:", git_prefix)


def print_changed_files(files):
    print('Changed files')
    for i in files:
        print('  ', i)


def print_changed_projects(changed_projects):
    print('Changed projects')
    for i in changed_projects:
        print('  ', i)


def print_project_info(project_infos):
    print('Project info')
    for k, v in project_infos.items():
        v = v.copy()
        v.pop('tools')
        print('  ', k, v)


def changed_files(config, circleci, commit, verbose):
    from menhir.gitutils import changed_files, repo, uncommited_files
    git_repo = repo(config['git_root'])
    uncommited = uncommited_files(git_repo)['all']
    if uncommited and not commit:
        files = uncommited
        if verbose:
            print("Uncommited files")
    else:
        from_commit, to_commit = commit_range(
            commit, circleci, config['git_root'], verbose
        )
        if verbose:
            print("Commit range: %s..%s" % (from_commit, to_commit))

        files = changed_files(from_commit, to_commit)['all']

    return files


def commit_range(commit_arg, circleci, root, verbose):
    import os
    from menhir.gitutils import head_commit, repo
    if commit_arg:
        return commits_from_range(commit_arg)
    if circleci:
        compare_url = os.getenv('CIRCLE_COMPARE_URL')
        log.debug('Deriving commit range from %s', compare_url)
        if compare_url:
            commit_range = compare_url.rsplit('/', 1)[1]
            return commits_from_range(commit_range)
        sha = os.getenv('CIRCLE_SHA1')
        return sha+"^", sha

    r = repo(root)
    commit = head_commit(r)
    return str(commit)+"^", str(commit)


def commits_from_range(commit_range):
    log.debug('commits_from_range: %s', commit_range)
    commits = commit_range.split('..')
    if len(commits) == 2:
        from_commit = commits[0].replace('.', '')
        to_commit = commits[1].replace('.', '')
        log.debug('commits_from_range: from %s to %s', from_commit, to_commit)
        return from_commit, to_commit
    log.debug('commits_from_range: single commit %s', commit_range)
    return commit_range+"^", commit_range



# GIT utilities

from git import Repo


def find_root():
    """Based on the current directory, look for a .menhir_root.yaml file."""
    from os.path import dirname
    from .fileutils import find_file_in_parents
    root = find_file_in_parents(".git")
    if root is None:
        raise FileNotFoundError(
            'No git repository in parent directories',
            ".git"
        )
    return dirname(root)


def repo(base_dir=None):
    """Return a repository object."""
    base_dir = base_dir or find_root()
    return Repo(base_dir)


def commit(repo, commitish):
    """Return a commit object."""
    return repo.commit(commitish)


def head_commit(repo):
    """Return a commit object."""
    return repo.head.commit


def diff(start, end):
    """Diff for the commit range."""
    return diff_paths(start.diff(end))


def dirty_files(repo):
    return diff_paths(repo.index.diff(None))


def staged_files(repo):
    return diff_paths(repo.index.diff("HEAD"))


def uncommited_files(repo):
    dirty = repo.index.diff(None)
    dirty.extend(repo.index.diff("HEAD"))
    return diff_paths(dirty)


def diff_paths(diff):
    rpaths = [p.a_path for p in diff.iter_change_type('R')]
    rpaths.extend([p.b_path for p in diff.iter_change_type('R')])

    diffs = {
        'added': [p.b_path for p in diff.iter_change_type('A')],
        'deleted': [p.a_path for p in diff.iter_change_type('D')],
        'renamed': rpaths,
        'modified': [p.b_path for p in diff.iter_change_type('M')],
    }
    all_paths = []
    for i in ['added', 'deleted', 'renamed', 'modified']:
        all_paths.extend(diffs[i])
    diffs['all'] = all_paths
    return diffs


def changed_files(start, end):
    from menhir import gitutils
    repo = gitutils.repo()
    start_commit = gitutils.commit(repo, start)
    end_commit = gitutils.commit(repo, end)
    changed_files = gitutils.diff(start_commit, end_commit)
    return changed_files



# pyenv based tool for menhir
import logging

import menhir.tool as mtool
from menhir.tool_utils import OK, FAIL, NOTHING_TO_DO, tool_env, working_dir

log = logging.getLogger(__name__)


class Pyenv():
    pass


def tool():
    return Pyenv()


@mtool.name.register(Pyenv)
def name(arg):
    return "pyenv"


@mtool.is_using_tool.register(Pyenv)
def is_using_tool(tool, path):
    from os.path import exists, join
    path = join(path, 'requirements.txt')
    return exists(path)


@mtool.dependencies.register(Pyenv)
def dependencies(tool, path):
    return []


def run(cmd, *args, **kwargs):
    import subprocess
    res = subprocess.call(cmd, *args, **kwargs)
    if res:
        return False, FAIL
    return True, OK


@mtool.execute_build_phase.register(Pyenv)
def execute_build_phase(
        tool, path, info, phase_name, args,
        only_if_changed=False,
        verbose=False,
):
    """Execute a build phase."""
    from os.path import basename
    import subprocess
    from menhir.tool_utils import package_script

    if phase_name == 'requirements':
        if (
                info.get('changed') or
                info.get('changed_dependents') or
                (info.get('changed_dependees') and not only_if_changed)
        ):
            if verbose:
                print('Running pyenv requirements in %s' % path)
            log.debug('pyenv requirements in %s', path)

            env = tool_env()

            with package_script("/tools/pyenv/requirements.sh") as f:
                with working_dir(path):
                    res = subprocess.call(
                        [f.name, path, basename(path)],
                        env=env,
                    )
                    if res:
                        return FAIL
                    return OK
        else:
            if verbose:
                print('Not running pyenv requirements in %s' % path)
            log.debug('not running pyenv requirements in %s', path)
            return OK
    else:
        return NOTHING_TO_DO
